#include "mea_evaluator.hh"
#include "mea_predictor.hh"
#include "LocARNA/confusion_matrix.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <cassert>

//typedef MeaSpanPenaltyPredictor<MeaCLPredictor> predictor_t;
typedef MeaSpanPenaltyPredictor<MeaOSPredictor> predictor_t;
typedef MeaMixDPsPredictor<MeaOSPredictor> predictor_mix_t;

MeaEvaluator::MeaEvaluator(std::istream &in, const std::string &dpdir,
			   const bool slide_rule,
			   const bool conflict_rule,
			   const LocARNA::ConfusionMatrix::BPFilter *bpfilter
			   )
    : slide_rule_(slide_rule),
      conflict_rule_(conflict_rule),
      bpfilter_(bpfilter)  { 
    read_fasta_with_structures(in, reference_ );
    read_corresponding_dotplots(reference_, dpdir, dotplots_);
}

MeaEvaluator::MeaEvaluator(const std::string &filename,
			   const std::string &dpdir,
			   const std::string &pred,
			   const bool slide_rule,
			   const bool conflict_rule,
			   const LocARNA::ConfusionMatrix::BPFilter *bpfilter
			   ) 

    : slide_rule_(slide_rule),
      conflict_rule_(conflict_rule),
      bpfilter_(bpfilter)  { 

    std::ifstream in(filename.c_str());
    
    if (! in.good()) {
	throw std::ios_base::failure("Cannot open "+filename+" for reading.");
    }

    if(! dpdir.empty())
	{
	    read_fasta_with_structures(in, reference_ );
	    read_corresponding_dotplots(reference_, dpdir, dotplots_);
	}
    else
	{
	    std::ifstream in2(pred.c_str());
	    if (! in2.good()) {
		throw std::ios_base::failure("Cannot open "+pred+" for reading.");
	    }
	    init(in,in2);
	    in2.close();
	}
    in.close();
}




void
MeaEvaluator::read_fasta_with_structures(std::istream &in, 
					 structure_map_t &structure_map)
{
    while (!in.eof())
	{
	    std::string sLine;
	    std::string name = "";
	    std::getline(in,sLine);
      
	    if(sLine.substr(0,1)==">")
		{
		    name = sLine.substr(1,sLine.size()-1);
		    std::getline(in,sLine);
		    std::string sequence = sLine;
		    std::getline(in,sLine);
		
		    size_t firstblank=sLine.find(" ");
		    if (firstblank != std::string::npos) {
			sLine=sLine.substr(0,firstblank);
		    }
		    structure_map[name] = sLine;
		    
		    if (structure_map[name].length() != sequence.length()) {
			std::cerr << "Length of sequence and structure differ!"<<std::endl;
			std::cerr << sequence <<std::endl;
			std::cerr << structure_map[name].to_string()<<std::endl;
		    }
		}
	}
}


void
MeaEvaluator::read_corresponding_dotplots(const structure_map_t &structure_map, 
					  const std::string &dpdir, 
					  dotplots_t &dotplots)
{
    for(structure_map_t::const_iterator it=structure_map.begin();
	structure_map.end() != it; ++it)
	{	
	    const std::string &name = it->first;
	    
	    //read dot plot file for <name>
	    
	    dotplots_t::iterator dp_it=
		dotplots.insert(std::pair<std::string,dotplot_t >
				(name,dotplot_t(0))).first;
	    // we can't write "dotplots[name] =
	    // dotplot_t(0);" if there is no default
	    // constructor for dotplot_t
	    
	    std::string path = dpdir + "/" + name + "_dp.ps";
	    std::ifstream ins(path.c_str());
	    
	    if (! ins.good()) {
		throw std::ios_base::failure("Cannot open "+path+" for reading.");
	    }
	    
	    predictor_t::readDotPlot(ins,sequences_[name],dp_it->second);
	    
	    ins.close();
	}
}


void
MeaEvaluator::init(std::istream &ref, std::istream &pred) {
    while (!ref.eof())
	{
	    std::string sLine;
	    std::string name = "";
	    std::getline(ref,sLine);
      
	    if(sLine.substr(0,1)==">")
		{
		    name = sLine.substr(1,sLine.size()-1);
		    getline(ref,sLine);
		    std::string sequence = sLine;
		    getline(ref,sLine);
		
		    size_t firstblank=sLine.find(" ");
		    if (firstblank != std::string::npos) {
			sLine=sLine.substr(0,firstblank);
		    }
		    reference_[name] = sLine;

		    if (reference_[name].length() != sequence.length()) {
			std::cerr << "Length of sequence and reference structure differ!"<<std::endl;
			std::cerr << sequence <<std::endl;
			std::cerr << reference_[name].to_string()<<std::endl;
		    }

		}
	}

    while (!pred.eof())
	{
	    std::string sLine;
	    std::string name = "";
	    std::getline(pred,sLine);
      
	    if(sLine.substr(0,1)==">")
		{
		    name = sLine.substr(1,sLine.size()-1);
		    std::getline(pred,sLine);
		    std::string sequence = sLine;
		    std::getline(pred,sLine);
		
		    size_t firstblank=sLine.find(" ");
		    if (firstblank != std::string::npos) {
			sLine=sLine.substr(0,firstblank);
		    }
		    predictions_[name] = sLine;

		    if (predictions_[name].length() != sequence.length()) {
			std::cerr << "Length of sequence and reference structure differ!"<<std::endl;
			std::cerr << sequence <<std::endl;
			std::cerr << predictions_[name].to_string()<<std::endl;
		    }
		}
	}
}

double MeaEvaluator::evaluate(double alpha,double beta,double gamma,double delta) const {
    double f_score=0.0;
    double cf_score=0.0;

    if(predictions_.empty()) {  //used dotplots to get structures

	for (dotplots_t::const_iterator it=dotplots_.begin(); dotplots_.end()!=it; ++it ) {
	    std::string name =it->first;
	    const dotplot_t &dp=it->second;
	    //use mea_predictor to get pred structure
    
	    const std::string &sequence = sequences_.find(name)->second;
	    const LocARNA::RnaStructure &reference = reference_.find(name)->second;
    
	    predictor_t mp(sequence,dp, alpha,beta,gamma,delta);
    
	    mp.fold();
	    std::string predicted_structure = mp.traceback();

	    LocARNA::ConfusionMatrix cm(reference,
					LocARNA::RnaStructure(predicted_structure),
					slide_rule_,
					conflict_rule_,
					bpfilter());

	    // std::cout << ">" << name << std::endl
	    // 	      << reference.to_string() << std::endl
	    // 	      << predicted_structure << " \t("
	    // 	      << cm.f1_score() << ")" << std::endl
	    // 	      << "#" << cm.tp() << " " << cm.fp() << " "
	    // 	      << cm.fn() << " "  << cm.tn() << " "  
	    // 	      << cm.f1_score() << " "  << cm.mcc() << " "
	    // 	      << predicted_structure.length() << std::endl;
    
	    f_score += cm.f1_score();
	}
  
	//std::cerr << "CF1_score: "<<cf_score/(double)reference_.size()<<std::endl;
	return f_score/(double)dotplots_.size();

    }


    else {  //got file with already predicted structures

	for (structure_map_t::const_iterator it=reference_.begin(); reference_.end()!=it; ++it ) {
	    std::string name = it->first;
	    if ( predictions_.find (name) == predictions_.end() ) {
		std::cout << "Structures in references found which are not in predictions!" << std::endl;
	    }
	    else {
		const LocARNA::RnaStructure &refstruct = it->second;
		const LocARNA::RnaStructure &predstruct = predictions_.find(name)->second;
		LocARNA::ConfusionMatrix cm(refstruct,
					    predstruct,
					    slide_rule_,
					    conflict_rule_,
					    bpfilter());

		// std::cout << ">" << name << std::endl
		// 	  << refstruct.to_string() << std::endl
		// 	  << predstruct.to_string() << " \t("
		// 	  << cm.f1_score() << ")" << std::endl
		// 	  << "#" << cm.tp() << " " << cm.fp() << " "
		// 	  << cm.fn() << " "  << cm.tn() << " "  
		// 	  << cm.f1_score() << " "  << cm.mcc() << " "
		// 	  << predstruct.length() << std::endl;
		
		f_score  += cm.f1_score();
	    }
	}

	return f_score/(double)reference_.size();

    }

}

MeaMixEvaluator::MeaMixEvaluator(std::istream &in,
				 const std::string &dpdir1,
				 const std::string &dpdir2,
				 const bool slide_rule,
				 const bool conflict_rule,
				 const LocARNA::ConfusionMatrix::BPFilter *bpfilter) 
    : MeaEvaluator(in,dpdir1,slide_rule,conflict_rule,bpfilter)
{
    read_corresponding_dotplots(reference_, dpdir2, dotplots2_);
}


double
MeaMixEvaluator::evaluate(double gamma1,double gamma2) const {

    double f_score=0.0;

    for (structure_map_t::const_iterator it=reference_.begin(); reference_.end()!=it; ++it ) {
	
	const std::string &name                = it->first;
	const LocARNA::RnaStructure &reference = it->second;

	assert(dotplots_. find(name) != dotplots_.end());
	const dotplot_t &dp  = dotplots_. find(name)->second;
	
	assert(dotplots2_. find(name) != dotplots2_.end());
	const dotplot_t &dp2 = dotplots2_.find(name)->second;
	
	//use mea_predictor to get pred structure
	
	assert(sequence_. find(name) != sequence_.end());
	const std::string &sequence = sequences_.find(name)->second;
	
	predictor_mix_t mp(sequence, dp, dp2, gamma1, gamma2);
	
	mp.fold();

	std::string predicted_structure = mp.traceback();
	
	LocARNA::ConfusionMatrix cm(reference,LocARNA::RnaStructure(predicted_structure),
				    slide_rule_,
				    conflict_rule_,
				    bpfilter());
	
	f_score += cm.f1_score();
    }
    
    return f_score/(double)dotplots_.size();
}
