#ifndef MEA_EVALUATOR_HH
#define MEA_EVALUATOR_HH

#include <tr1/unordered_map>

#include "LocARNA/sparse_matrix.hh"
#include "LocARNA/rna_structure.hh"
#include "LocARNA/confusion_matrix.hh"
#include "mea_predictor.hh"

/**
 * @brief Evaluate mea prediction over benchmark set
 */
class MeaEvaluator {

protected:
    const bool slide_rule_;
    const bool conflict_rule_;
    
public:
    //! hash of reference structures by name
    typedef std::tr1::unordered_map<std::string, LocARNA::RnaStructure> structure_map_t; 
    //! triangle matrix of scaled probabilities
    typedef LocARNA::SparseMatrix<double> dotplot_t;
    
    //! hash of dotplots by name
    typedef std::tr1::unordered_map<std::string, dotplot_t> dotplots_t;
    
    //! hash of sequences by name
    typedef std::tr1::unordered_map<std::string, std::string> sequences_t;


    /** 
     * @brief Construct with list of instances and directory of dot plots
     * 
     * @param in input stream containing names, sequences and reference structures
     * @param dpdir directory with dotplots, names in in file
     * @see init
     */
    MeaEvaluator(std::istream &in,
		 const std::string &dpdir,
		 const bool slide_rule,
		 const bool conflict_rule,
		 const LocARNA::ConfusionMatrix::BPFilter *bpfilter
		 );

    /** 
     * @brief Construct with list of instances and directory of dot plots, and either param dpdir or pred
     *
     * @param in filename of file with names, sequences and reference structures
     * @param dpdir directory with dotplots, names in in file
     * @param pred file with already predicted structures to compare with references,
     * @see init
     */ 
    MeaEvaluator(const std::string &filename,
		 const std::string &dpdir, 
		 const std::string &pred,
		 const bool slide_rule,
		 const bool conflict_rule,
		 const LocARNA::ConfusionMatrix::BPFilter *bpfilter
		 );
    
    /** 
     * @brief evaluate at specific parameters
     * 
     * @param alpha alpha parameter of base pair penalty function 
     * @param beta  beta parameter of base pair penalty function 
     * @param gamma weight of base pair score over unpaired score 
     * @param delta delta parameter of base pair penalty function 
     * 
     * @return average f1-score over benchmark set 
     */
    double
    evaluate(double alpha,double beta,double gamma,double delta) const;


    void 
    set_bpfilter(const LocARNA::ConfusionMatrix::BPFilter *bpfilter) {
	bpfilter_ = bpfilter;
    }
    
protected:

    /**
     * @brief Read sequences and structures from a fasta-like input stream
     *
     * @param in input stream containing names, sequences and reference structures
     * @param structure_map[out] structures of sequences by name
     *
     * Read in benchmark RNAs with reference structures from input stream
     * in fasta-like format 
     *   >name
     *   sequence
     *   structure as dot bracket string.
     */
    void
    read_fasta_with_structures(std::istream &in, structure_map_t &structure_map);
    
    /**
     *
     * @param dpdir directory with dotplots, names in in file
     *
     * Read in dot plots from files dot plot directory; names in the
     * fasta file have to be file names of dot plot ps files in this directory.
     */
    void
    read_corresponding_dotplots(const structure_map_t &structure_map, 
				const std::string &dpdir, 
				dotplots_t &dotplots);
    
    void
    init(std::istream &ref, std::istream &pred);
    
    const LocARNA::ConfusionMatrix::BPFilter &
    bpfilter() const {
	return *bpfilter_;
    }

    sequences_t sequences_;
    
    structure_map_t reference_;    

    //! hash of predicted structures by name
    structure_map_t predictions_; 

    dotplots_t dotplots_;
private:

    const LocARNA::ConfusionMatrix::BPFilter *bpfilter_;
};


class MeaMixEvaluator : public MeaEvaluator {
public:
    /** 
     * @brief Construct with list of instances and two directories of dot plots
     * 
     * @param in input stream containing names, sequences and reference structures
     * @param dpdir1 first directory with dotplots, names in in file
     * @param dpdir2 second directory with dotplots, names in in file
     * @see init
     */
    MeaMixEvaluator(std::istream &in,
		    const std::string &dpdir1,
		    const std::string &dpdir2,
		    const bool slide_rule,
		    const bool conflict_rule,
		    const LocARNA::ConfusionMatrix::BPFilter *bpfilter);

    /** 
     * @brief evaluate at specific parameters
     * 
     * @param alpha alpha parameter of base pair penalty function 
     * @param beta  beta parameter of base pair penalty function 
     * @param gamma weight of base pair score over unpaired score 
     * @param delta delta parameter of base pair penalty function 
     * 
     * @return average f1-score over benchmark set 
     */
    double
    evaluate(double gamma1,double gamma2) const;
    
    //! hash of dotplots by name
    std::tr1::unordered_map<std::string, dotplot_t> dotplots2_;
    
};

#endif
