#ifndef MEAPREDICTOR_H
#define MEAPREDICTOR_H

#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "LocARNA/sparse_matrix.hh"

/**
 * @brief MEA RNA secondary structure predictor
 */
class MeaPredictorBase {
protected:
  typedef int score_t; //!< type of internal scores

public:
  typedef LocARNA::SparseMatrix<double> pair_probs_t;
  typedef std::vector<double> unpaired_probs_t;

  MeaPredictorBase(std::istream &in, double gamma) 
    : PSCALE_(10000),
      min_loop_size_(3),
      gamma_(gamma),
      p_(0)
  {
    readDotPlot(in,sequence_,p_);
    length_=sequence_.length();
    comp_unpaired_probs(p_,pu_);
  }

  MeaPredictorBase(const std::string &sequence, const pair_probs_t &p, double gamma) 
    : PSCALE_(10000),
      min_loop_size_(3),
      sequence_(sequence),
      p_(p),
      gamma_(gamma),
      length_(sequence.length())
  {
    comp_unpaired_probs(p_,pu_);
  }
  
  /**
   * @brief optimize Mea score
   * @return maximum score
   */
  virtual
  double
  fold() = 0;

  /**
   * @brief obtain best structure by traceback
   * @return best structure
   */
  virtual
  std::string 
  traceback() = 0;

  /**
   * @brief compute base unpairing probabilities
   *
   */
  void 
  comp_unpaired_probs(const pair_probs_t &p, unpaired_probs_t &pu);

  /**
   * @brief access sequence
   */
  const std::string &sequence() const { return sequence_;}
  
  /**
   * @brief read data from dotplot file
   */
  static
  void
  readDotPlot(std::istream &in, std::string &sequence, pair_probs_t& p);

private:
  const int PSCALE_; //!< scaling of probabilities
  const size_t min_loop_size_; //!< minimal loop size

  size_t length_; //!< sequence length
  std::string sequence_;//!< sequence

  double gamma_; //!< gamma parameter
  
protected:  
  int PSCALE() const {return PSCALE_;}
  int min_loop_size() const {return min_loop_size_;}
  double gamma() const {return gamma_;}
  size_t length() const {return length_;}
  
  /**
   * @brief Base pair score
   *
   * @param i left end
   * @param j right end
   * @return score of base pair (i,j)
   */
  virtual
  score_t
  bp_score(size_t i, size_t j) const {
    return 2*gamma_*p_(i,j)*PSCALE();
  }
  
  /**
   * @brief Unpaired score
   *
   * @param i left end
   * @return score of unpaired base i
   */
  virtual
  score_t
  unpaired_score(size_t i) const {
    assert(0<=i && i<length_);
    return pu_[i]*PSCALE();
  }

  pair_probs_t p_;           //!< base pair probabilities
  unpaired_probs_t pu_;          //!< base unpair probabilities
  

}; // end class MeaPredictorBase

/**
 * @brief mea predictor with candidate-list sparsification
 */
class MeaCLPredictor : public MeaPredictorBase {
 public:
  
  MeaCLPredictor(std::istream &in, double gamma) 
    : MeaPredictorBase(in,gamma)
  {
  }

  MeaCLPredictor(const std::string &sequence, const pair_probs_t &p, double gamma) 
    : MeaPredictorBase(sequence,p,gamma)
  {
  }
  
  std::string
  traceback();

  double
  fold();

  void
  print_matrix_debug() const;
  
protected:
  std::vector<std::vector<int> > dp_; //!< dynamic programming matrix 
  
  /**
   * @brief traceback after DPfold, write structure_ std::vector
   * @param i base location in sequence
   * @param j base location in sequence
   */
  void
  traceback(int i, int j);

private:
  std::string structure_;
  
}; // end class MeaCLPredictor


/**
 * @brief mea predictor with oct-step time and space efficient sparsification
 */
class MeaOSPredictor : public MeaPredictorBase {
protected:
  typedef std::vector<std::pair<size_t, score_t> > oct_vector_t;

  typedef std::vector<oct_vector_t> octs_t;
public:
  
  MeaOSPredictor(std::istream &in, double gamma) 
    : MeaPredictorBase(in,gamma)
  {
  }

  MeaOSPredictor(const std::string &sequence, const pair_probs_t &p, double gamma) 
    : MeaPredictorBase(sequence,p,gamma)
  {
  }
  
  double
  fold();

  std::string
  traceback();

 protected:

  octs_t octs_; //!< octs and their scores
  
  std::vector<score_t> row_i_;   //!< row i
  std::vector<score_t> row_ip1_; //!< row i+1;

  
  /**
   * @brief traceback for subsequence i..j
   */
  void
  traceback(size_t i, size_t j);

  /**
   * @brief restore entries in column j for rows i..j
   */
  void
  restore(size_t i,size_t j);
  
  /**
   * @brief lookup j in octs_[i]
   * @return index of j in octs_[i] (or index of closest element if not found)
   */
  size_t
  lookup_oct(size_t i, size_t j, size_t from, size_t to) const {
    assert(i<=j);
    assert(j<length());
    assert(from<=to); // precondition
    assert(to<octs_[i].size());
    if (from==to) {
      return from;
    }
    
    size_t mid = (from+to)/2;
    
    if (octs_[i][mid].first >= j) {
      return lookup_oct(i,j,from,mid);
    } else {
      return lookup_oct(i,j,mid+1,to);
    }
  }
  
  /**
   * @brief test whether i..j is oct
   */
  bool
  is_oct(size_t i,size_t j) const {
    if (octs_[i].size()==0) return false;
    
    size_t idx = lookup_oct(i,j,0,octs_[i].size()-1);
    bool ans = octs_[i][idx].first == j;
    return ans;
  }
  
  /**
   * @brief get score of oct i..j (pre: i..j is oct)
   */
  score_t
  get_oct_score(size_t i,size_t j) const {
    size_t idx=lookup_oct(i,j,0,octs_[i].size()-1);
    assert(octs_[i][idx].first == j); // precondition
    return octs_[i][idx].second;
  }

  void 
  print_matrix_debug() const;
  
private:
  std::string structure_;

}; // end class MeaOSPredictor


/**
 * @brief MEA RNA secondary structure predictor with span-dependant penalties
 */
template<class parent_t>
class MeaSpanPenaltyPredictor : public parent_t
{
protected:
  typedef typename parent_t::score_t score_t;
  
  /**
   * @brief Base pair penalty
   *
   * @param i left end
   * @param j right end
   * @return penalty factor for base pair (i,j)
   */
  double
  penalty(size_t i, size_t j) const {
    assert(j>=i);
    size_t d = j-i;
    // double delta_prime=delta_/(2.0*parent_t::gamma());
    // return delta_prime+(1.0-delta_prime)/(1.0+exp(-alpha_*(-(double)d+beta_)));
    return delta_+(1.0-delta_)/(1.0+exp(-alpha_*(-(double)d+beta_)));
  }
  
  /**
   * @brief Base pair score
   *
   * @param i left end
   * @param j right end
   * @return score of base pair (i,j)
   */
  score_t
  bp_score(size_t i, size_t j) const {
    return 
      penalty(i,j) * parent_t::bp_score(i,j);
  }
  
public:

  /**
   * @brief construct using dotplot filename
   * @param filename of dotplot input file
   */
  MeaSpanPenaltyPredictor(std::istream &in,double alpha,double beta,double gamma,double delta)
    :
    parent_t(in,gamma),
    alpha_(alpha), 
    beta_(beta),
    delta_(delta)
  {
  }
  
  MeaSpanPenaltyPredictor(const std::string sequence, const LocARNA::SparseMatrix<double>& p,double alpha,double beta,double gamma,double delta)
    : parent_t(sequence,p,gamma),		
      alpha_(alpha), beta_(beta), delta_(delta)
  {
  }
  
  void
  print_penalty(unsigned int lim) {
    for(unsigned int i=4;i<=lim;i++)
      {
	std::cout <<i<<"\t"<<penalty(0,i)<<"\t"<<2*parent_t::gamma()*penalty(0,i)<<std::endl;
      }
  }

protected:  

  double alpha_; //!< alpha parameter
  double beta_;  //!< beta parameter
  double delta_; //!< delta parameter

};

/**
 * @brief MEA RNA secondary structure predictor that mixes in
 * a second dot plot for scoring base pairs
 */
template <class parent_t>
class MeaMixDPsPredictor : public parent_t
{
  //  typedef MeaCLPredictor parent_t;
protected:
  typedef typename parent_t::score_t score_t;
  typedef typename parent_t::pair_probs_t pair_probs_t;
  typedef typename parent_t::unpaired_probs_t unpaired_probs_t;
  
  /**
   * @brief Base pair score
   *
   * @param i left end
   * @param j right end
   * @return score of base pair (i,j)
   */
  score_t
  bp_score(size_t i, size_t j) const {
    score_t s= parent_t::bp_score(i,j) + 2.0 * gamma2_ * p2_(i,j) * parent_t::PSCALE();
    return s;
  }

  score_t
  unpaired_score(size_t i) const {
    assert(0<=i && i<parent_t::length());
    return parent_t::unpaired_score(i) + (gamma2_/parent_t::gamma()*pu2_[i])*parent_t::PSCALE();
  }
  
public:
  /**
   * @brief construct using dotplot filename
   * @param filename of dotplot input file
   */
  MeaMixDPsPredictor(std::istream &in,std::istream &in2, double gamma, double gamma2)
    :
    parent_t(in,gamma),
    gamma2_(gamma2),
    p2_(0)
  {
    std::string sequence2;
    parent_t::readDotPlot(in2,sequence2,p2_);
    if (sequence2!=parent_t::sequence()) {
      std::cerr << "Warning: unequal sequences in dot plots." << std::endl;
    }
    if (sequence2.length()!=parent_t::sequence().length()) {
      std::cerr << "ERROR: sequences in dot plots have unequal length." << std::endl;
      throw std::exception();
    }

    parent_t::comp_unpaired_probs(p2_,pu2_);

  }
  
  MeaMixDPsPredictor(const std::string sequence, const LocARNA::SparseMatrix<double>& p, const LocARNA::SparseMatrix<double>& p2, double gamma, double gamma2)
    : parent_t(sequence,p,gamma),
      p2_(p2),
      gamma2_(gamma2)
  {
    parent_t::comp_unpaired_probs(p2_,pu2_);
  }

private:
  pair_probs_t p2_;  //!< pair probabilities 2
  unpaired_probs_t pu2_;          //!< base unpair probabilities 2
  double gamma2_;  //!< gamma parameter for second pair probability matrix
};

#endif //MEAPREDICTOR_H
